import numpy as np
import torch
from transformers import BertTokenizer
from tools import save_pickle
from textrank4zh import TextRank4Sentence
from gensim.models import KeyedVectors
from collections import Counter


class DataEncodePrecess(object):
    def __init__(self, tokenizer_file, word2vce_file):
        super(DataEncodePrecess, self).__init__()

        self.tokenizer = BertTokenizer.from_pretrained(tokenizer_file)
        self.word2vce = KeyedVectors.load_word2vec_format(word2vce_file,
                                                          binary=False,
                                                          encoding="utf-8",
                                                          unicode_errors='ignore')

    def bert_encode(self, bert_text):
        words = self.tokenizer.encode_plus(text=bert_text)
        return {"input_ids": words['input_ids'], "attention_mask": words['attention_mask']}

    def rank_encode(self, lstm_text):
        result = ''
        for x in lstm_text:
            result += x
        tr4s = TextRank4Sentence()
        tr4s.analyze(result, lower=True, source='all_filters')
        sentences = tr4s.get_key_sentences(num=1, sentence_min_len=1)
        sentences = [x for x in sentences[0]['sentence']]
        key_sentence = self.lstm_encode(lstm_text=sentences)
        return key_sentence

    def lstm_encode(self, lstm_text):
        words = []
        for word in lstm_text:
            if word in self.word2vce:
                words.append(self.word2vce[word].tolist())
            else:
                words.append([0] * 200)
        return words


def read_data_save_pickle(input_file, tokenizer_file, word2vce_file):
    feature_encode = DataEncodePrecess(
        tokenizer_file=tokenizer_file,
        word2vce_file=word2vce_file)

    with open(input_file, 'r', encoding='utf-8') as f:
        out_lists = []
        entries = f.read().strip().split("\n\n")
        all_labels = []
        for entry in entries:
            words = []
            ner_labels = []

            for line in entry.splitlines():
                pieces = line.strip().split()
                if len(pieces) < 2:
                    continue
                words.append(pieces[0])
                ner_labels.append(pieces[-1])

            bert_encode_dict = feature_encode.bert_encode(bert_text=words)
            lstm_encode = feature_encode.lstm_encode(lstm_text=words)
            rank_encode = feature_encode.rank_encode(lstm_text=words)

            bert_encode_dict['rnn_array'] = lstm_encode
            # bert_encode_dict['rank_array'] = rank_encode

            predict_mask = [0] + [1] * len(ner_labels)

            bert_encode_dict['ner_labels'] = ['[CLS]'] + ner_labels + ['[SEP]']
            bert_encode_dict['predict_mask'] = predict_mask
            out_lists.append(bert_encode_dict)
            all_labels += ner_labels
    return out_lists, all_labels


def save():
    data_list_train, all_labels_train = read_data_save_pickle(input_file=r'medical.train',
                                                              tokenizer_file=r'../bert-base-chinese',
                                                              word2vce_file=r'word_vce.txt')
    save_pickle(data_list_train, '../pickle_data/train_data.pickle')

    data_list_valid, all_labels_valid = read_data_save_pickle(input_file=r'medical.dev',
                                                              tokenizer_file=r'../bert-base-chinese',
                                                              word2vce_file=r'word_vce.txt')
    save_pickle(data_list_valid, '../pickle_data/train_valid.pickle')

    data_list_test, all_labels_test = read_data_save_pickle(input_file=r'medical.test',
                                                            tokenizer_file=r'../bert-base-chinese',
                                                            word2vce_file=r'word_vce.txt')
    save_pickle(data_list_test, '../pickle_data/train_test.pickle')

    print(f'Train Set Labels {Counter(all_labels_train)}')
    print(f'Valid Set Labels {Counter(all_labels_valid)}')
    print(f' Test Set Labels {Counter(all_labels_test)}')

    # Train Set Labels Counter({'O': 147864, 'I-西医诊断': 14731, 'I-临床表现': 6962, 'I-方剂': 4167, 'I-中药': 3655,
    # 'I-中医证候': 3480, 'I-中医治疗': 3338, 'B-临床表现': 3068, 'B-中药': 2786, 'B-西医诊断': 2444, 'I-西医治疗': 1646,
    # 'B-中医证候': 1170, 'B-中医治疗': 1081, 'B-方剂': 1012, 'I-中医治则': 714, 'B-西医治疗': 465, 'I-中医诊断': 325,
    # 'B-中医治则': 297, 'B-中医诊断': 246, 'I-其他治疗': 218, 'B-其他治疗': 65})

    # Valid Set Labels Counter({'O': 18412, 'I-西医诊断': 1901, 'I-临床表现': 815, 'I-中药': 553, 'I-方剂': 541,
    # 'B-中药': 429, 'I-中医证候': 408, 'B-临床表现': 369, 'I-中医治疗': 353, 'B-西医诊断': 308, 'I-西医治疗': 148,
    # 'I-中医治则': 142, 'B-中医证候': 136, 'B-方剂': 120, 'B-中医治疗': 111, 'B-中医治则': 53, 'I-中医诊断': 49,
    # 'B-西医治疗': 44, 'B-中医诊断': 42, 'I-其他治疗': 21, 'B-其他治疗': 8})

    #  Test Set Labels Counter({'O': 17624, 'I-西医诊断': 1985, 'I-临床表现': 810, 'I-中医证候': 476, 'I-方剂': 437,
    #  'I-中医治疗': 398, 'B-临床表现': 375, 'B-西医诊断': 330, 'I-中药': 328, 'B-中药': 262, 'I-西医治疗': 219,
    #  'B-中医证候': 158, 'B-中医治疗': 133, 'I-中医治则': 122, 'B-方剂': 111, 'I-中医诊断': 55, 'B-西医治疗': 53,
    #  'B-中医诊断': 48, 'B-中医治则': 47, 'I-其他治疗': 40, 'B-其他治疗': 12})


if __name__ == '__main__':
    save()
